import sys
import random
import datetime
import traceback
from datetime import timedelta

from ardi.driver import histdriverbase
import MySQLdb

def fixdate(dt):
	"""Used to fix the incoming date format on some requests."""
	ps = dt.find('.')
	if ps > -1:
		return dt[0:ps]
	return dt

class ardihistconnectordemo:
    """An example historical connector

    In this case, the connector is for a MySQL database. Note that any member variables of this class are actually private -
    the system only requires that the FUNCTIONS are implemented.

    Historian connectors MUST implement the SetAddress, Connect, Disconnect and Query functions. """
    def __init__(self):
        """Initialises the connector """

        ##A reference to the core, which contains data and functions that the connector requires.
        self.core = None

        ##The database port. 0 for the MySQL default
        self.port = 0

        ##The database connection instance
        self.db = None

        ##The database host name
        self.host = ""

        ##The database name
        self.database = ""

        ##The login user name
        self.username = ""

        ##The login password
        self.password = ""

        ##The query to send to the data source
        self.query = ""
        
        ##The field that represents the date stamp
        self.stamp = ""

        ##The table to query
        self.table = ""

        ##The field that will match the lookup value connecting ARDI to the data source
        self.lookup = ""

        ##The field that contains the value
        self.value = ""

        ##An SQL filter string
        self.filter = ""

        ##A lookup table to accelerate data lookups
        self.lookuptable = {}

        ##A flag used when retrying a query due to a connection failure
        self.retry = False

    def SetAddress(self, addr):
        """Set the address of the connection target

        This is a string that contains all of the details required to connect to the MySQL database.

        \param addr: A colon (:) delimited list of connection properties """
        self.connected = False
        bits = addr.split(":")
        filter = ""

        #print "Raw Address"

        if len(bits) >= 6:
            self.host = bits[0]
            self.port = bits[1]
            self.database = bits[2]
            self.username = bits[3]
            self.password = bits[4]
            self.table = bits[5]
            self.lookup = bits[6]
            self.stamp = bits[7]
            self.value = bits[8]
        if len(bits) > 9:
            filter = bits[9]

    def Connect(self):
        """Connect to the datasource

        \return Returns false on a connection error, True on success """
        try:
        	print "Connecting!"
        	self.db = MySQLdb.connect(self.host,self.username,self.password,self.database)
        	self.connected = True        	
        	
        	return True
        except:
        	return False			

    def Disconnect(self):
        """Disconnect from the datasource """
        try:			
        	self.con.close()
        except:
        	pass	

    def Query(self, points, start,end,function,grain,first):
		"""Performs a database query.

		Note that this function is NOT trivial.

		The query must also include those points immediately outside and (if 'First' is True) immediately before
		the range.

		It must also summarise data according to the 'function' and 'grain' parameters

		The query must return those 'outside' points to ensure that the data is properly filled in.

		The outside points do not need to be truncated or interpolated - ARDI will take care of that
		on it's end.

		\param points: A list of points to query
		\param start: The start date for the query, in YYYY-MM-DD hh:mm:ss format
		\param end: The start date for the query, in YYYY-MM-DD hh:mm:ss format
		\param function: The function to perform. Functions include 'raw', 'avg', 'sum', 'min' and 'max'
		\param grain: The number of seconds between samples, or '0' for natural sampling rate
		\param first: True if this the data is to include the samples immediately preceeding the start time.
		"""
		sd = datetime.datetime.strptime( start, "%Y-%m-%d %H:%M:%S" )
		ed = datetime.datetime.strptime( end, "%Y-%m-%d %H:%M:%S" )

		#OK - fabricate an SQL query...			
		func = "AVG"
		if function == "min":
			func = "MIN"
		if function == "max":
			func = "MAX"

		x = 0
		lookuptable = {}
		for p in points:			
			lookuptable[str(p)] = x
			x = x + 1

		out = ""

		try:
			self.db.commit()
		except:
			pass

		#print points

		if function == "raw":			
			if first == 1:
				topoffquery = "SELECT " + self.stamp + ", " + self.lookup + "," + self.value +" FROM ( SELECT * "		
				topoffquery += "FROM " + self.table + " WHERE " + self.stamp + " <= '" + str(sd) + "' AND " + self.lookup + " IN ("

				x=0
				for p in points:
					if x > 0:
						topoffquery += ","
					topoffquery += "'" + p + "'"
					x = x + 1

				topoffquery += ") ORDER BY " + self.stamp + " DESC) AS t1 GROUP BY " + self.lookup
				
				csr = self.db.cursor()
				csr.execute(topoffquery);
				
				#This is the first fetch of raw data. As such, I'm going to need the PREVIOUS points for interpolation...

				for row in csr:
					out += str(lookuptable[str(row[1])]) + "," + str(row[0]) + "," + str(row[2]) + "\r\n"

				csr.close()
		
		self.query = "SELECT " + self.stamp + "," + self.lookup + "," + self.value + " FROM " + self.table + " WHERE "
		if grain != 0:
			self.query = "SELECT DATE_ADD('" + str(sd) + "',INTERVAL (CEIL(TIMESTAMPDIFF(SECOND,'" + str(sd) + "'," + self.stamp + ") / " + str(grain) + ") * " + str(grain) + ") SECOND) as ostamp, " + self.lookup + " as lookup,"+ func + "(" + self.value + ") as value" + " FROM " + self.table + " WHERE "
			
		if self.filter != "":
			self.query += "( " + self.filter + " ) AND "
			
		self.query += "( " + self.stamp + " >= '" + str(sd) + "' AND " + self.stamp + " <= '" + str(ed) + "' )"

		#Filter for point list...
		self.query += " AND " + self.lookup + " IN ( "
		x = 0		
		for p in points:
			if x > 0:
				self.query += ","
			self.query += "'" + p + "'"			
			x = x + 1

		#print lookuptable

		self.query += " )"
		if grain != 0:
			self.query += " GROUP BY ostamp, lookup ORDER BY ostamp,lookup"
		else:
			self.query += " ORDER BY " + self.stamp + " ASC"

		#print "Submitting Query: " + self.query
		
		try:			
			csr = self.db.cursor()			

			csr.execute(self.query)		
			for row in csr:
				out += str(lookuptable[str(row[1])]) + "," + str(row[0]) + "," + str(row[2]) + "\r\n"

			csr.close()
		except:
			if self.retry == false:
				self.retry = True
				try:
					self.db = MySQLdb.connect(self.host,self.username,self.password,self.database)
					self.connected = True
					return self.query(points, start,end,function,grain,first)
				except:
					pass
			return ""
		
		topoffquery = "SELECT " + self.stamp + ", " + self.lookup + "," + self.value +" FROM ( SELECT * "		
		topoffquery += "FROM " + self.table + " WHERE " + self.stamp + " >= '" + str(ed) + "' AND " + self.lookup + " IN ("

		x=0
		for p in points:
			if x > 0:
				topoffquery += ","
			topoffquery += "'" + p + "'"
			x = x + 1

		topoffquery += ") ORDER BY " + self.stamp + " ASC) AS t1 GROUP BY " + self.lookup

		try:	
			csr = self.db.cursor()
			csr.execute(topoffquery);
		
			#OK - same deal, but this time get the NEXT row value from each...

			for row in csr:
				out += str(lookuptable[str(row[1])]) + "," + str(row[0]) + "," + str(row[2]) + "\r\n"

			csr.close()
		except:
			pass

		try:
			self.db.commit()
		except:
			pass

		#print out

		self.retry = False

		return out.encode('ascii')
 
class ardihistconnectorfactorydemo:
    """Demo historian connector factory

    Every driver (live or historian) requires a 'Connector Factory' - a class that produces new instances of the
    class that manages a connection to a data source.

    It must implement the 'createinstance' function. """
    def createinstance(self):
        """Create a single connector instance """
        return ardihistconnectordemo()
            
if __name__ == "__main__":
    sdf = ardihistconnectorfactorydemo()
    base = histdriverbase.historian()
    base.start(sdf)