import pycurl
import xmltodict
import time
import socket
import transform
import platform
from subprocess import Popen
import subprocess
import logging
import logging.handlers
import thread
import os
import pytz
from tzlocal import get_localzone
import datetime
from datetime import timedelta
import traceback
import math

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from StringIO import StringIO

class pointinfo:
    """Data Point Information

    An empty class used for storing information about data points"""
    pass

class historianquery2:
	def __init__(self,engine,points,start,end,function,grain,first):

		self.core = engine
		self.function = function
		self.lastvalues = {}
		self.lasttimestamps = {}
		self.grain = grain
		self.ends = end
		self.starts = start
		self.first = first
		self.points = points
		self.tz = "UTC"
		self.datestyle = "1"

		self.output = ""
		self.quotestring = False

		self.addrbindings = {}
		self.addresses = []

		self.localtz = pytz.UTC

		self.sd = datetime.datetime.strptime( start, "%Y-%m-%d %H:%M:%S" )
		self.ed = datetime.datetime.strptime( end, "%Y-%m-%d %H:%M:%S" )

		self.localstart = self.sd
		self.localend = self.ed

		if self.function == "min" or self.function == "max":
		    self.calchannels = []
		    
		if self.function == "avg":
		    self.totals = []
		    self.counters = []

		qryindx = -1
		for pnt in points:
		    qryindx = qryindx+1
		    plindex = -1
		    for itm in self.core.pointlist:
		        plindex = plindex + 1
		        if itm.code == pnt:
		            if itm.address not in self.addresses:
		                self.addresses.append(itm.address)
		                self.addrbindings[itm.address] = []
		            self.addrbindings[itm.address].append((qryindx,plindex))
		            if self.function == "min":
		                self.calchannels[plindex] = 999999.9
		            if self.function == "max":
		                self.calchannels[plindex] = -999999.9
		            if self.function == "avg":
		                self.counters[plindex] = 0
		                self.totals[plindex] = 0

	def Cleanup(self):
	    vv = ""
	    stamp = datetime.datetime.strptime( self.ends, "%Y-%m-%d %H:%M:%S" )
	    if (self.function == "changes"):
	        #In 'Changes' mode, we need to output any resudual values...
	        for keys in self.lasttimestamps:
	            if self.lasttimestamps[keys] != False:
	                vv += str(keys) + "," + str(stamp) + "," + str(self.lastvalues[keys]) + "\r\n"

	    self.output += vv

	def DTTextToNative(self,dt,tzoverride = ""):
	    #Traditional date...
	    #self.core.logger.info("Converting " + str(dt) + " To Native")
	    tz = self.tz
	    localtz = self.localtz

	    if (tzoverride != ""):
	        tz = tzoverride
	        if tz == "local":
	            localtz = get_localzone()
	        else:
	            if tz == "UTC":
	                localtz = pytz.UTC
	            else:
	                localtz = pytz.timezone(self.tz)

	    if self.datestyle == "1":
	        if tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC).replace(tzinfo=None)
	        if self.quotestring==True:
	            return "'" + str(dt) + "'";
	        else:
	            return str(dt);

	    #UNIX Epoch
	    if self.datestyle == "2":
	        if tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        return str(long((dt - datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC)).total_seconds()))

	    #Win32 FILETIME
	    if self.datestyle == "3":
	        tickspersec = 10000000
	        if self.tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        l = long((dt - datetime.datetime(1601,1,1,0,0,0,0,pytz.UTC)).total_seconds())
	        return str(l * tickspersec);
	    
	    #CLR Time
	    if self.datestyle == "4":
	        tickspersec = 10000000
	        if self.tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        diff = (dt - datetime.datetime(1,1,1,0,0,0,0,pytz.UTC)).total_seconds()
	        return str(long(tickspersec) * long(diff));

	def DTUTCToNative(self,dt,tzoverride = ""):
	    #Traditional date...
	    #self.core.logger.info("Converting " + str(dt) + " To Native")
	    tz = self.tz
	    localtz = self.localtz

	    if (tzoverride != ""):
	        tz = tzoverride
	        if tz == "local":
	            localtz = get_localzone()
	        else:
	            if tz == "UTC":
	                localtz = pytz.UTC
	            else:
	                localtz = pytz.timezone(self.tz)

	    if tz != "UTC":
	        dt = pytz.UTC.localize(dt)
	        dt = dt.astimezone(localtz).replace(tzinfo=None)

	    if self.datestyle == "2":
	        if tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        return str(long((dt - datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC)).total_seconds()))

	    #Win32 FILETIME
	    if self.datestyle == "3":
	        tickspersec = 10000000
	        if self.tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        l = long((dt - datetime.datetime(1601,1,1,0,0,0,0,pytz.UTC)).total_seconds())
	        return str(l * tickspersec);
	    
	    #CLR Time
	    if self.datestyle == "4":
	        tickspersec = 10000000
	        if self.tz != "UTC":
	            dt = localtz.localize(dt)
	            dt = dt.astimezone(pytz.UTC)
	        diff = (dt - datetime.datetime(1,1,1,0,0,0,0)).total_seconds()
	        return str(long(tickspersec) * long(diff));  

	    if self.quotestring==True:
	        return "'" + str(dt) + "'";
	    else:
	        return str(dt);

	def DTNativeToLocalDate(self,dt):
	    #Traditional date...
	    if self.datestyle == "1":
	        return dt
	
	    #UNIX Epoch
	    if self.datestyle == "2":
	    	dd = datetime.datetime.utcfromtimestamp(long(dt)).replace(tzinfo=pytz.UTC)
	    	return dd.astimezone(self.localtz).replace(tzinfo=None)
	    
	    #Win32 FILETIME
	    if self.datestyle == "3":
	        tickspersec = 10000000
	        l = long(dt) / tickspersec
	        dvx = (datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC) - datetime.datetime(1601,1,1,0,0,0,0,pytz.UTC)).total_seconds()
	        l -= dvx
	        #print "Final Timestamp: " + str(l)
	        dd = datetime.datetime.utcfromtimestamp(l).replace(tzinfo=pytz.UTC)
	        return dd.astimezone(self.localtz).replace(tzinfo=None)

	    #CLR Time
	    if self.datestyle == "4":
	    	tickspersec = 10000000
	    	diff = long(dt) / tickspersec
	    	tstamp = datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC)
	    	dvx = (datetime.datetime(1970,1,1,0,0,0) - datetime.datetime(1,1,1,0,0,0)).total_seconds()
	    	#print "Total Seconds: " + str(long(dvx))
	    	dd = datetime.datetime.utcfromtimestamp(diff - dvx).replace(tzinfo=pytz.UTC)
	    	return dd.astimezone(self.localtz).replace(tzinfo=None)

	def DTNativeToUTCDate(self,dt):
	    #Traditional date...
	    #print("Incoming: " + str(dt))
	    if self.datestyle == "1":
	        try:
	            if self.tz != "UTC":
	                dt = self.localtz.localize(dt)
	                #print "Ummm... localised, its " + str(dt)
	                dt = dt.astimezone(pytz.UTC).replace(tzinfo=None)
	                #print "Final Result: " + str(dt)
	            return dt
	        except:
	            ps = dt.find('.')
	            if ps > -1:
	                dt = dt[0:ps]
	            dt = datetime.datetime.strptime( dt, "%Y-%m-%d %H:%M:%S" )
	            if self.tz != "UTC":
	                dt = self.localtz.localize(dt)
	                #print "Ummm... lcalised, its " + str(dt)
	                dt = dt.astimezone(pytz.UTC).replace(tzinfo=None)
	                #print "Final Result: " + str(dt)
	        return dt

	    #UNIX Epoch
	    if self.datestyle == "2":
	    	dd = datetime.datetime.utcfromtimestamp(long(dt)).replace(tzinfo=pytz.UTC)
	    	return dd.replace(tzinfo=None)

	    #Win32 FILETIME
	    if self.datestyle == "3":
	        tickspersec = 10000000
	        l = long(dt) / tickspersec
	        dvx = (datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC) - datetime.datetime(1601,1,1,0,0,0,0,pytz.UTC)).total_seconds()
	        l -= dvx
	        #print "Final Timestamp: " + str(l)
	        dd = datetime.datetime.utcfromtimestamp(l).replace(tzinfo=pytz.UTC)
	        return dd.replace(tzinfo=None)

	    #CLR Time
	    if self.datestyle == "4":
	    	tickspersec = 10000000
	    	diff = long(dt) / tickspersec
	    	tstamp = datetime.datetime(1970,1,1,0,0,0,0,pytz.UTC)
	    	dvx = (datetime.datetime(1970,1,1,0,0,0) - datetime.datetime(1,1,1,0,0,0)).total_seconds()
	    	#print "Total Seconds: " + str(long(dvx))
	    	dd = datetime.datetime.utcfromtimestamp(diff - dvx).replace(tzinfo=pytz.UTC)
	    	return dd.replace(tzinfo=None)

	def SetDateFormat(self,timezone,style=1,quotestrings=False):
	    self.datestyle = style
	    self.tz = timezone
	    if self.tz == "local":
	        self.localtz = get_localzone()
	    else:
	        if self.tz == "UTC":
	            self.localtz = pytz.UTC
	        else:
	            self.localtz = pytz.timezone(self.tz)

	    self.quotestring = quotestrings
	    self.sd = self.DTUTCToNative(self.sd)
	    self.ed = self.DTUTCToNative(self.ed)
	    self.core.logger.info("Range: " + str(self.sd) + " - " + str(self.ed) + " ( UTC - From " + timezone + " )")

	def Finish(self):
	    self.Cleanup()
	    self.core.logger.info("Query Complete")
	    return self.output.encode('ascii')

	def AddLine(self,addr,stamp,value):
	    #Convert the stamp to UTC and output it.
	    stamp = self.DTNativeToUTCDate(stamp)
	    try:
	        bindings = self.addrbindings[addr]
	        for bind in bindings:
	            vx = value
	            #print addr + " Data Is Bound To " + self.core.pointlist[bind[1]].code
	            id = bind[0]
	            #Apply any transformations or corrections to the data here.
	            if self.core.pointlist[bind[1]].transform != "":
	                tx = transform.transform(self.core.pointlist[bind[1]].transform)
	                #print "Transforming: " + self.core.pointlist[bind[1]].transform
	                vx = tx.run(vx)

	            if self.core.pointlist[bind[1]].discrete == True:
	                #print(self.core.pointlist[bind[1]].code + " is discrete!")
	                vx = int(vx)

	            #This section handles any requests that are only looking for changes in your data
	            if (self.function == "changes"):
	                if id not in self.lastvalues:
	                    self.lastvalues[id] = vx
	                    self.lasttimestamps[id] = False
	                else:
	                    if (self.lastvalues[id] == vx):
	                        self.lasttimestamps[id] = stamp
	                        return
	                    else:
	                        oldstamp = self.lasttimestamps[id]
	                        self.lasttimestamps[id] = False
	                        oldvalue = self.lastvalues[id]
	                        self.lastvalues[id] = vx
	                            
	                        if oldstamp != False:
	                            self.output += str(id) + "," + str(oldstamp) + "," + str(oldvalue) + "\r\n" + str(id) + "," + str(stamp) + "," + str(vx) + "\r\n"
	                            return
	                        
	            if (self.function == "min"):
	                try:
	                    f = float(vx)
	                    if f < self.calchannels[id]:
	                        self.calchannels[id] = f
	                    return
	                except:
	                    pass
	            if (self.function == "max"):
	                try:
	                    f = float(vx)
	                    if f < self.calchannels[id]:
	                        self.calchannels[id] = f
	                    return
	                except:
	                    pass
	            if (self.function == "avg"):
	                try:
	                    f = float(vx)
	                    self.counters[id] = self.counters[id] + 1
	                    self.totals[id] += f
	                    return
	                except:
	                    pass
	                            
	            self.output += str(id) + "," + str(stamp) + "," + str(vx) + "\r\n"
	    except:
	        #traceback.print_exc()
	        return

class historianquery:
	def __init__(self,cr,fun,endtime):
		self.core = cr
		self.function = fun
		self.lastvalues = {}
		self.lasttimestamps = {}
		self.grain = 2
		self.ends = endtime

	def Cleanup(self,stamp):
	    vv = ""
	    if (self.function == "changes"):
	        #In 'Changes' mode, we need to output any resudual values...
	        for keys in self.lasttimestamps:
	            if self.lasttimestamps[keys] != False:
	                vv += str(keys) + "," + str(stamp) + "," + str(self.lastvalues[keys]) + "\r\n"

	    return vv

	def AddRecord(self, id, stamp, value, addr):
		if self.core.transforms[addr] != "":
		    tx = transform.transform(self.core.transforms[addr])
		    value = tx.run(value)

		if (self.function == "changes"):
			if id not in self.lastvalues:
				self.lastvalues[id] = value
				self.lasttimestamps[id] = False
			else:
				#print "Comparing Against " + str(value)
				if (self.lastvalues[id] == value):
					self.lasttimestamps[id] = stamp
					#print "Skipping Line!"
					return ""
				else:	
					oldstamp = self.lasttimestamps[id]
					self.lasttimestamps[id] = False
					oldvalue = self.lastvalues[id]
					self.lastvalues[id] = value
						
					if oldstamp != False:
						return str(id) + "," + str(oldstamp) + "," + str(oldvalue) + "\r\n" + str(id) + "," + str(stamp) + "," + str(value) + "\r\n"

		return str(id) + "," + str(stamp) + "," + str(value) + "\r\n"
	
class historiancore:
    """A connection to a single instance or device

    Unlike live drivers, historian drivers are mostly passive and don't run their own individual threads.

    They wait until they are queried by the system to return data.
    """
    
    def __init__(self,driver):
        """ Initialises the Core"""
        ##The url to return data to, used for remote drivers
        self.url = "localhost"

        ##The source ID
        self.asset = 1

        ##The actual driver instance - the class that performs the query on the data store
        self.driver = driver

        ##A self-reference        
        driver.core = self

        ##A number to track any dynamic addresses (usually associated with random data)
        driver.dynaddress = 0

        self.qry = historianquery(self,'raw',"")
    
    def SetARDIDetails(self, url, assetid):
        """Set Connection Details

        This function sets some of the basic identity information about the core.

        Args:
            url: The URL to query to get core configuration information from ARDI
            assetid: The source ID that this core represents"""    
        self.url = url
        self.asset = assetid
        self.InitLogger()
    
    def SetARDIDriver(self, driver):
        """Set Driver

        This function associates the given driver with this core

        Args:
            driver: Associate this core with a driver"""       
        self.driver = driver
        self.driver.asset = self.asset
        driver.core = self
                 
    def Close(self):
		"""Closes to connection to the data store"""
		self.driver.Disconnect()

    def InitLogger(self):
        """Initialises the logging for this indiviual source"""

        if hasattr(self,'logger'):
            return
        
        #print 'Initialising Logger for ' + self.asset
        logfile = '/var/log/ardi/source-' + self.asset + '-' + self.profile + '-hist.log'
        if (platform.system() == 'Windows'):
                    logfile = "c:\\windows\\temp\\ardi-source-" + self.asset + '-' + self.profile + '-hist.log'

        ##The logger for a specific data source, rather than the driver process as a whole
        self.logger = logging.getLogger('Source '+self.asset)  
        if (len(self.logger.handlers) == 0):
            handler = logging.handlers.WatchedFileHandler(logfile)
            handler.setLevel(logging.INFO)
            formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.info('----DRIVER STARTED----')
            self.logger.info('Starting Historian Data Source ' + self.asset + " on thread " + str(thread.get_ident()))
	
    def Reload(self):
        """Force the core to reload

        This function is usually called after a change to the required data"""
        self.LoadConfig()
        
    def LoadConfig(self):
        """Reload driver configuration

        This function triggers the driver to reload it's details from the ARDI server. Note that this
        is very different from the same function located in the DriverBase class. The data returned by
        this query includes all of the individual data points to be sampled from the data source."""
        if (self.source != ''):
			fl = open(self.source,'r')
			xml = fl.read()
			fl.close()
			self.parseconfig(xml)
			return
			
        response = StringIO()
        
        fullurl = "http://" + self.url + "/api/sourcedata?asset=" + str(self.asset) + "&mode=1"
        
        c = pycurl.Curl()
        c.setopt(c.URL,fullurl)
        c.setopt(c.WRITEFUNCTION,response.write)
        c.setopt(c.TIMEOUT,5)
        c.setopt(c.CONNECTTIMEOUT,5)
        c.perform()
        c.close()
        
        self.parseconfig(response.getvalue())
        
    def parseconfig(self,xml):
        """Parse the configuration details

        This function processes the XML from 'LoadConfig' to setup the driver and the many data points."""
        #print xml
        result = xmltodict.parse(xml)
        
        #print str(result)
        
        #Core Parameters
        self.driveraddr = result['sourcedata']['connection']['address']

        self.hostport = int(result['sourcedata']['connection']['port'])
        self.driver.SetAddress(self.driveraddr)

        if self.driveraddr is None:
            self.driveraddr = ""

        #Data Points
        #try:
        if True:
            self.pointlist = []
            self.transforms = {}
            if result['sourcedata']['points'] is not None:
                pointset = result['sourcedata']['points']['point']
                if '@address' in pointset:
                    #There is a Single Point
                    pointset = [pointset]
                
                for point in pointset:
                    #Add Data Points to List
                    if point['@address'] == '':
                        point['@address'] = point['@code']

                    self.transforms[point['@address']] = point['@transform']
                    
                    newpoint = pointinfo()
                    newpoint.address = point['@address']
                    newpoint.code = point['@code']
                    newpoint.changed = False
                    newpoint.value = ""
                    newpoint.intvalue = ""
                    newpoint.transform = point['@transform']
                    newpoint.discrete = False
                    try:
                        vv = newpoint.code.index(':value')
                        newpoint.discrete = True
                    except:
                        pass
                    try:
                        vv = newpoint.code.index(':state')
                        newpoint.discrete = True
                    except:
                        pass
                    self.pointlist.append(newpoint)
        
        self.logger.info(str(len(self.transforms)) + " Points/Transforms Loaded")

        try:
            self.driver.Remap()
        except:
            pass
    
    def Query(self, points,start,end,function,grain,first):		
        """Performs a query on a single source

        \param points: A list of points to query
        \param start: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param end: The start date for the query, in YYYY-MM-DD hh:mm:ss format
        \param function: The function to perform. Functions include 'raw', 'avg', 'sum', 'min' and 'max'
        \param grain: The number of seconds between samples, or '0' for natural sampling rate
        \param first: True if this the data is to include the samples immediately preceeding the start time.
		"""

        if hasattr(self.driver, 'RunQuery') == False:
            self.logger.warning("Driver Using Old (v1) Calls")
        else:
            try:
                #print str(dir(self.driver))
                hq = historianquery2(self,points,start,end,function,grain,first)
                return self.driver.RunQuery(hq)
            except:
                self.logger.exception("Historical Query (v2) Failed")
                return None
        
        finalpoints = []
        for pnt in points:
            for itm in self.pointlist:				
                if itm.code == pnt:
                    if itm.address not in finalpoints:
                          finalpoints.append(itm.address)
                    break

        if len(finalpoints) == 0:
            return "ERROR: No valid point specified"
        try:
            return self.driver.Query(finalpoints,start,end,function,grain,first)
        except:
            self.logger.exception("Historical Query Failed")
            return None

    def GetResponder(self,func,endtime):
    	return historianquery(self,func,endtime)

    def AddRecord(self, id, stamp, value,addr):
        return self.qry.AddRecord(id,stamp,value,addr)

    def Launch(self,address,asset):
        """Starts the driver

        This function initiates the driver.

        Args:
            address: The encoded destination address (ie. IP address) of the device
            asset: The asset id that this source represents"""
        self.SetARDIDetails(address,asset)
        self.LoadConfig()    
        self.driver.Connect()
                    