import SocketServer
try:
    from urllib.parse import parse_qsl
except:
    from urlparse import parse_qsl

from threading import Thread

def QueryFunction(socket,inst,cmd):
    print "Query Thread Started"
    socket.Respond(inst,cmd)
    print "Query Thread Stopped"

class IPCHandler(SocketServer.StreamRequestHandler):
    """ Historian IPC Message Handler

    This class responds to commands that are issued from server applications such as ARDI.

    Some commands operate on the driver as a whole, while others operate on a single core """
    def handle(self):
        """ Handles Incoming Commands

        This function reads from the incoming stream and response appropriately."""
        self.data = self.rfile.readline().strip()
        cmd = self.data.split('|')
        #print cmd        
        try:            
            if cmd[0] == 'spinup':
                if len(cmd) > 1:
                    self.server.core.Spinup(cmd[1])
                    return
            if cmd[0] == 'reload':
                if len(cmd) > 1:
                    self.server.core.Reload(cmd[1])
                    return
            if cmd[0] == 'spindown':
                self.server.core.Restart()
                return
            if cmd[0] == 'restart':
                self.server.core.Restart()
                return
            if cmd[0] == 'query':
                #Rebuild query...
                str = ""
                for x in range(2,len(cmd)):
                    if x > 2:
                        str = str + "|"
                    str = cmd[x]
                    
                self.ThreadRespond(cmd[1],str)
                return
        except:
            pass

        if self.data[0:14] == "GET /query?id=":
            #Get the ID number...
            pos = self.data.find("&")
            if pos > 0:
                id = self.data[14:pos]            
                self.ThreadRespond(id,self.data[pos+1:len(self.data)-9])

    def ThreadRespond(self,instance,cmd):
        self.Respond(instance,cmd)
        #reactor.callInThread(QueryFunction,self,instance,cmd)

    def Respond(self,instance,cmd):
        """ Sends a query response to the client

        This function issues a query via the driver and returns the results.
       
        \param instance The ARDI source asset ID
        \param cmd The query command string"""
        #Split Query String

        vals = parse_qsl(cmd)
        
        points = ""
        start = ""
        end = ""
        func = "raw"
        grain = 0
        first = 1
        
        params = cmd.split("&")
        for x in vals:            
            if x[0] == "points":
                points = x[1].split("_")
            if x[0] == "start":
                start = x[1]
            if x[0] == "end":
                end = x[1]
            if x[0] == "func":
                func = x[1]
            if x[0] == "function":
                func = x[1]
            if x[0] == "grain":
                grain = x[1]
            if x[0] == "followup":
                first = 0

        #print "Requesting History..."
        result = self.server.core.Query(instance,points,start,end,func,grain,first)
        #print "Sending Response: " + result
        self.wfile.write(result)

class ThreadedTCPServer(SocketServer.ThreadingMixIn, SocketServer.TCPServer):
    pass

class DriverIPC:
    """ Historical IPC (Socket) server to control a driver

    This class listens for socket connections and responds to commands from controlling applications (such as ARDI)

    Some commands operate on the driver as a whole, while others operate on a single core """
    def __init__(self):
        """Intialises the IPC Connection"""

        ##The TCPServer instance that represents this service
        self.service = None

        ##The port that this IPC service should run on
        self.port = 9001

        ##A reference to the drivercore that spawned this IPC service
        self.core = None        

    def start(self):
        """Start the IPC Service

        This command starts the TCPServer that listens for incoming commands"""
        SocketServer.TCPServer.allow_reuse_address = 1
        self.service = SocketServer.TCPServer(("",self.port),IPCHandler)
        #self.service = ThreadedTCPServer(("",self.port),IPCHandler)
        self.service.core = self.core
        self.service.serve_forever()    

#IPC = DriverIPC()
#IPC.start()