import SocketServer

class IPCHandler(SocketServer.StreamRequestHandler):
    """ Live IPC Message Handler

    This class responds to commands that are issued from server applications such as ARDI.

    Some commands operate on the driver as a whole, while others operate on a single core """
    def handle(self):
        """ Handles Incoming Commands

        This function reads from the incoming stream and response appropriately."""

        ##A single incoming command
        #print "Processing IPC Message"
        self.data = self.rfile.readline().strip()
        cmd = self.data.split('|')
        print cmd
        try:
            if cmd[0] == 'pause':
                if len(cmd) > 1:
                    self.server.core.Pause(cmd[1])
            if cmd[0] == 'resume':
                if len(cmd) > 1:
                    self.server.core.Resume(cmd[1])
            if cmd[0] == 'spinup':
                if len(cmd) > 1:
                    self.server.core.Spinup(cmd[1])
            if cmd[0] == 'reload':
                if len(cmd) > 1:
                    self.server.core.Reload(cmd[1])
            if cmd[0] == 'uprate':
                if len(cmd) > 2:
                    self.server.core.SetUpdateRate(cmd[1],cmd[2])                
            if cmd[0] == 'spindown':
                self.server.core.Restart()
            if cmd[0] == 'restart':
                self.server.core.Restart()
            if cmd[0] == 'debug':
                self.server.core.DebugMode(cmd[1],cmd[2])
        except Exception as e:
            print "Exeption Processing Command: " + str(e)
            pass

class DriverIPC:
    """ Live IPC (Socket) server to control a driver

    This class listens for socket connections and responds to commands from controlling applications (such as ARDI)

    Some commands operate on the driver as a whole, while others operate on a single core """
    def __init__(self):
        """Intialises the IPC Connection"""

        ##The TCPServer instance that represents this service
        self.service = None

        ##The port that this IPC service should run on
        self.port = 9001

        ##A reference to the drivercore that spawned this IPC service
        self.core = None
        pass

    def start(self):
        """Start the IPC Service

        This command starts the TCPServer that listens for incoming commands"""
        SocketServer.TCPServer.allow_reuse_address = 1
        self.service = SocketServer.TCPServer(("",self.port),IPCHandler)
        self.service.core = self.core
        self.service.serve_forever()    

#IPC = DriverIPC()
#IPC.start()